//
// JHLClassDependancy.java
//
// Copyright (c) 1996 John Leach. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for NON-COMMERCIAL purposes and without fee is hereby granted
// provided that this copyright notice appears in all copies.
//
// JOHN LEACH MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
// THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
// NON-INFRINGEMENT. JOHN LEACH SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
// BY A LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
// OR ITS DERIVATIVES.
//

package apps.jclass;

import java.io.*;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;

import jclass.*;

/**
 * This application compiles a dependancy list for a given <tt>.class</tt> file.
 * See <a href="../apps/jclass/ClassDependancy.html">A <tt>.class</tt>ic utility</a>
 * for further details.
 *
 * @version	1.0, 27 June 1996.
 * @author	<a href="mailto:leachj@easy1.easynet.it">John Leach</a>
**/

public class JHLClassDependancy {

 /** The system defined class path. */
 String classPath;

 /** The system file separator. */
 String fileSeparator;

 /** The system path separator. */
 String pathSeparator;

 /**
  * Starts the application.
  * @param args[]   the command line arguments
 **/

 public static void main(String args[]) {
  JHLClassDependancy compiler = new JHLClassDependancy();
  compiler.compile(args);
  }

 /**
  * Creates an initialised compiler.
 **/

 public JHLClassDependancy() {
  Properties prop = System.getProperties();
  classPath = prop.getProperty("java.class.path");
  fileSeparator = prop.getProperty("file.separator");
  pathSeparator = prop.getProperty("path.separator");
  }

 /**
  * Parses the <tt>.class</tt> file, and compiles the dependancy list
  * @param args[]   the command line arguments
 **/

 public void compile(String args[]) {
  int argc = 0;
  String userClassPath = null;
  String classFile = null;

  try {
   while(argc < args.length) {
    if(args[argc].startsWith("-")) {
     if(args[argc].equals("-classpath")) userClassPath = args[++argc];
     else {
      System.out.println("Unknown argument '" + args[argc] + "'");
      showUsage();
      return;
      }
     }
    else {
     if(classFile != null) {
      System.out.println("Only one .class file may be specified");
      showUsage();
      return;
      }
     classFile = args[argc];
     }
    argc++;
    }
   }
  catch(ArrayIndexOutOfBoundsException e) {
   showUsage();
   return;
   }

  // check for compulsory input

  if(classFile == null) {
   System.out.println("No .class file specified");
   showUsage();
   return;
   }

  // split up the classpath into the various individual paths

  if(userClassPath == null) userClassPath = classPath;
  if(classPath == null || pathSeparator == null || fileSeparator == null) {
   System.out.println("Unable to retrieve Java system properties");
   return;
   }

  Vector classPaths = new Vector();
  StringTokenizer tok = new StringTokenizer(userClassPath, pathSeparator);
  while(tok.hasMoreTokens()) {
   String path = tok.nextToken();
   if(path.endsWith(fileSeparator)) path = path.substring(0, path.length() - 1);
   classPaths.addElement(path);
   }

  // process the top level file

  if(classFile.endsWith(".class")) classFile = classFile.substring(0, classFile.length() - 6);
  if(classFile.startsWith(fileSeparator)) classFile = classFile.substring(1);
  classFile = classFile.replace('.', fileSeparator.charAt(0));
  classFile = classFile + ".class";
  boolean done = false;
  String fileName = ""; // otherwise we get a compiler error
  ClassData classData = new ClassData();

  Enumeration enum = classPaths.elements();
  while(enum.hasMoreElements() & !done) {
   fileName = (String)enum.nextElement() + fileSeparator + classFile;
   InputStream dataIn = null;
   try{
    try {
     dataIn = new FileInputStream(fileName);
     // System.out.println("Parsing '" + classFile + "'");
     new Parser(dataIn, classData);
     done = true;
     }
    finally { if(dataIn != null) dataIn.close(); }
    }
   catch(FileNotFoundException e) { continue; }
   catch(Exception e) {
    System.out.println("Exception occurred in file '" + fileName + "'");
    System.out.println(e.getMessage());
    return;
    }
   }

  // check that we parsed the file

  if(!done) {
   System.out.println("File '" + classFile + "' not found");
   showUsage();
   return;
   }
  classData.setFileName(fileName);

  // add this .class file to the list

  Hashtable dependancy = new Hashtable();
  dependancy.put(fileName, classData);

  // iterate through the dependancies

  addDependants(classPaths, dependancy, classData);

  // print the dependancy list

  System.out.println("\nDependancy list for class file '" + classFile + "':");

  printOrderedList(dependancy);
  }

 /**
  * Parses all the dependancies of this <tt>.class</tt> file.
  * @param path     the separated classpath paths
  * @param parsed   the hash table of already parsed classes
  * @param data     the parsed .class file
 **/

 protected void addDependants(Vector path, Hashtable parsed, ClassData data) {

  // first the super class

  String name = data.getSuperClassName();
  name = name.replace('/', '.');

  if(!name.equals("") && !name.startsWith("java.") && !name.startsWith("sun.")) {
   ClassData superClass = parseClassFile(path, parsed, name);
   if(superClass != null) addDependants(path, parsed, superClass);
   }

  // then the used classes

  Vector used = data.getUsedClasses();
  if(used != null) {
   Enumeration enum = used.elements();
   while(enum.hasMoreElements()) {
    name = (String)enum.nextElement();
    name = name.replace('/', '.');

    if(name != null && !name.startsWith("java.") && !name.startsWith("sun.")) {
     ClassData usedClass = parseClassFile(path, parsed, name);
     if(usedClass != null) addDependants(path, parsed, usedClass);
     }
    }
   }
  }

 /**
  * Parses a <tt>.class</tt> file given the class name.
  * @param path     the separated classpath paths
  * @param parsed   the hash table of already parsed classes
  * @param name     the class name
  * @return	    the ClassData parsed if successful, otherwise null
 **/

 protected ClassData parseClassFile(Vector path, Hashtable parsed, String name) {
  String classFile = name.replace('.', fileSeparator.charAt(0));
  classFile = classFile + ".class";
  boolean done = false;
  Enumeration enum = path.elements();
  String fileName = ""; // otherwise we get a compiler error
  ClassData classData = new ClassData();

  while(enum.hasMoreElements() & !done) {
   fileName = (String)enum.nextElement() + fileSeparator + classFile;
   if(parsed.containsKey(fileName)) return null;
   InputStream dataIn = null;
   try{
    try {
     dataIn = new FileInputStream(fileName);
     // System.out.println("Parsing '" + classFile + "'");
     new Parser(dataIn, classData);
     done = true;
     }
    finally { if(dataIn != null) dataIn.close(); }
    }
   catch(FileNotFoundException e) { continue; }
   catch(Exception e) {
    System.out.println("Exception occurred in file '" + fileName + "'");
    System.out.println(e.getMessage());
    return null;
    }
   }

  // check that we parsed the file

  if(!done) {
   System.out.println("Dependant file '" + classFile + "' not found");
   return null;
   }
  classData.setFileName(fileName);

  // add this .class file to the list

  parsed.put(fileName, classData);
  return classData;
  }

 /**
  * Puts the dependancy list in alphabetical order, and then prints it out.
  * @param parsed   the dependancy hash table
 **/

 private void printOrderedList(Hashtable parsed) {
  Vector ordered = new Vector();
  Enumeration enum = parsed.elements();
  while(enum.hasMoreElements()) {
   ClassData data = (ClassData)enum.nextElement();
   addFilename(data.getFileName(), ordered);
   }

  enum = ordered.elements();
  while(enum.hasMoreElements()) {
   String fileName = (String)enum.nextElement();
   System.out.println(fileName);
   }
  }

 /**
  * Adds a filename to the ordered Vector.
  * @param fileName    the filename to add
  * @param vect        the Vector to add
 **/

 private void addFilename(String fileName, Vector vect) {
  int upper, lower, key, comparison;
  String elem;

  key = comparison = 0;
  if((upper = vect.size()) > 0) {
   lower = -1;
   while(true) {
    key = (upper + lower) >> 1;
    elem = (String)vect.elementAt(key);
    if((comparison = fileName.compareTo(elem)) == 0) {
     return;
     }
    if(comparison < 0) upper = key;
    else lower = key;
    if(lower >= (upper - 1)) break;
    }
   }
  if(comparison > 0) ++key;
  vect.insertElementAt(fileName, key);
  }

 /**
  * Shows the usage message.
 **/

 public void showUsage() {
  System.out.println(" Usage:   java apps.jclass.JHLClassDependancy [options] file");
  System.out.println(" Options: -classpath path  the classpath where the .class files are located");
  }
 }
